;;;;
;;;; ecl-readline.lisp
;;;;
;;;; Copyright (C) 2008 Jason Aquilina <jfa7600@yahoo.com.au>
;;;;
;;;; This program is free software; you can redistribute it and/or
;;;; modify it under the terms of the GNU General Public License
;;;; as published by the Free Software Foundation; either version 2
;;;; of the License, or (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;;;;

;;;;
;;;; Readline support for ECL
;;;;
;;;; Ideas and code stolen from:
;;;;    CL-READLINE by Nikodemus Siivola <nikodemus@random-state.net>
;;;;    SB-READLINE by Miles Egan <miles@caddr.com>
;;;;

(defpackage ecl-readline
  (:use :cl :uffi))
(in-package :ecl-readline)

(defvar *history-size* 100)

(defvar *history-file* 
  (concatenate 'string (si::getenv "HOME") "/.ecl-history"))

(defvar *prompt* "")

(defvar *newline* (string #\newline))

(ffi:clines "#include <stdio.h>")
(ffi:clines "#include <readline/readline.h>")
(load-foreign-library "readline" :system-library t)

(uffi:def-foreign-var ("rl_already_prompted" rl-already-prompted) :int nil)
(uffi:def-function ("readline" rl-readline) ((prompt :cstring)) :returning :cstring)
(uffi:def-function ("add_history" rl-add-history) ((line :cstring)) :returning :void)
(uffi:def-function ("write_history" rl-write-history) ((line :cstring)) :returning :int)
(uffi:def-function ("stifle_history" rl-stifle-history) ((size :int)) :returning :void)
(uffi:def-function ("read_history" rl-read-history) ((filename :cstring)) :returning :int)

(defclass input-stream ()
  ((in-buffer :initform (make-string 0))
   (in-index  :initform 0)
   (out-buffer :initform (make-array 0 :element-type 'character :adjustable t :fill-pointer t))))

(defmethod si::stream-input-p ((stream input-stream))
  t)

(defmethod si::stream-output-p ((stream input-stream))
  nil)

(defmethod si::stream-read-char ((stream input-stream))
  (ensure-stream-data stream)
  (with-slots (in-buffer in-index) stream
    (let ((c (char in-buffer in-index)))
      (incf in-index)
      c)))

(defmethod si::stream-unread-char ((stream input-stream) character)
  (with-slots (in-index) stream
    (if (> in-index 0)
	(decf in-index))))

(defmethod si::stream-listen ((stream input-stream))
  nil)

(defmethod si::stream-clear-input ((stream input-stream))
  nil)

(defmethod si::stream-close ((stream input-stream) &key abort)
  nil)

(defmethod si::stream-peek-char ((stream input-stream))
  (ensure-stream-data stream)
  (with-slots (in-buffer in-index) stream
    (char in-buffer in-index)))

(defun ensure-stream-data (stream)
  (with-slots (in-buffer in-index) stream
    (if (= in-index (length in-buffer))
	(setq in-buffer (readline)
	      in-index 0))))

;;;
;;; *** This is a Hack ***
;;;
;;; This code is taken from ECL's source src/lsp/top.lsp and is the code used to
;;; generate the ECL prompt. We let ECL output its own prompt and set the readline
;;; variable rl_already_prompted to 1 which tells readline it doesn't need to
;;; output a prompt. BUT the prompt must still be passed to readline so readline's
;;; redisplay functions can update the display properly.
;;;
(defun ecl-tpl-prompt ()
  (format nil "~A~V,,,'>A "
          (if (eq *package* (find-package 'user)) "" (package-name *package*))
          (- si::*tpl-level* si::*step-level* -1)
          ""))

(defun readline ()
  (let* ((c-line (rl-readline (ecl-tpl-prompt)))
	 (line (convert-from-cstring c-line)))
    (rl-add-history c-line)
    (free-cstring c-line)
    (rl-write-history *history-file*)
    (concatenate 'string line *newline*)))

(defun enable (&key ((:history-file filename) nil filename-p) ((:history-size size) nil size-p))
  (when filename-p
    (setq *history-file* filename))
  (when size-p
    (setq *history-size* size))
  (rl-read-history *history-file*)
  (rl-stifle-history *history-size*)
  (setq rl-already-prompted 1)
  (setq *standard-input* (make-instance 'input-stream))
  (setq *terminal-io* (make-two-way-stream *standard-input* (two-way-stream-output-stream *terminal-io*))))

;;;; End of File